// sir_cpp BYU SIR file format C++ include file
//
// This include file defines a C++ programing interface for the BYU SIR
// file format.  A data type which stores the key SIR header information
// is defined to permit easy transfer of the SIR information into the
// various I/O and geometry routines.  Function prototypes are also defined.
// 
// written 18 May 2002 by DGL at BYU + based on sir_ez
// 
// (c) 2002 by BYU MERS 

#ifndef SIRCPP_H
#define SIRCPP_H

extern "C" {

#include <stdio.h>

/* define important variable and function prototypes from
   low level C routines */

#include "sir3.h"
#include "sir_ez.h"
#include "dsir.h"

typedef sir_head sirheader;

#ifdef NOT_DEF
/* 
   function prototypes (from sir_ez.c) 
   note: struct sirheader must exactly match struct sir_head
*/

void sir_init_head(struct sirheader *head);
/* sets optional parameter flags to their default values.
   call before using struct sirheader */

int get_sir_head_file(FILE *imf, struct sirheader *head);
/* read sir file header given file pointer. Returns negative value
   on read error. */

int get_sir_head_name(char *fname, struct sirheader *head, FILE **imf);
/* read sir file header given file name.  Returns negative value
   on read/open error. */

float *sir_data_alloc(struct sirheader *head);
/* returns a pointer for a malloc'd storage array sized for storage
   of the sir image data */

int get_sir_data(FILE *imf, float *stval, struct sirheader *head);
/* read sir file image data into a float array given file pointer.   
   Returns negative value on read error. Closes file on successful read. */

int get_sir_data_block(FILE *imf, float *stval, struct sirheader *head,
		       int x1, int y1, int x2, int y2);
 /* read a block of the SIR image data from file into float array
       . file imf must already be open 
       . file imf is NOT closed after read 
     block is defined by the lower-left corner (x1,y1) and
     the upper-right corner (x2,y2) where 0 < x <= nsx and  0 < y <= nsy

     The output image has (x2-x1+1) by (y2-y1+1) pixels indexed according to
     image_index = (y-1) * (x2-x1+1) + x where 0 < y <= (y2-y1+1) and
     0 < x <= (x2-x1+1) and 1 < image_index <= (x2-x1+1) * (y2-y1+1).
     Obviously, it must be true that x2>=x1 and y2>=y1

     returns -5 on an input error, -1 on a file read, seek or other error.
     returns 0 on success
*/

int get_sir_data_byte(FILE *imf, char *stval, struct sirheader *head, float smin,
		      float smax, int bmin, int bmax);
/* read sir file image data into a byte array given a file pointer.
   The input image values are first converted to floats and then
   scaled to byte using 

   byte_value=(bmax-bmin)*(float_value-smin)/(smax-smin)+bmin

   Note: this technique generally results in loss of information */

int get_sir(char *fname, struct sirheader *head, float **stval);
/* opens sir file, reads header, malloc's image data storage,
   reads data and closes output file.  Returns negative value on
   open, read, or malloc error */

void print_sir_head(FILE *omf, struct sirheader *head);
/* summary print sir file header information to file pointer */

int put_sir(char *fname, struct sirheader *head, float *stval);
/* write sir file header and data to file.  Returns negative value
   on write error. */

int sir_latlon2pix(float alon, float alat, float *x, float *y, struct sirheader *head);
/* compute pixel location given lat,lon location.  Return pixel number 
   as non-negative value.  A negative value indicates pixel is outside of 
   the image. */

void sir_pix2latlon(float x, float y, float *alon, float *alat, struct sirheader *head);
/* compute lat, lon from pixel location.  Will work for values outside of 
   image for most projections */

int sir_pix(float x, float y, int *ix, int *iy, struct sirheader *head);
/* given a floating point x,y value, compute the integer pixel locations
   ix,iy and return the pixel index as a non-negative value in the
   range 0...nsx*nxy-1 consistent with C array indexing.
   A negative value indicates pixel is outside of the image. */

int sir_lex(int ix, int iy, struct sirheader *head);
/* given a pixel x,y value returns the pixel index as non-negative value.
   A negative value indicates pixel is outside of image. 
   If ix and iy are in range (1 <= x <= head.nsx) (1 <= y <= head.nsy)) 
   the lexicographic index of pixel (x,y) is 
   index = (y-1)*head.nsx + x-1 where 1 <= index <= head.nsx*head.nsy 
   The range index=0...nsx*nxy-1 is consistent with C array indexing. */

int isir_lex(int *ix, int *iy, int i, struct sirheader *head);
/* given a pixel index i, in the range i=0...nsx*nxy-1 consistent with 
   C array indexing, returns the pixel coordinates in SIR convention
   (1 <= ix <= head.nsx) (1 <= iy <= head.nsy)) 
   A negative return value indicates index i is outside of image, 
   Zero is returned if i is in the range (0..nsx*nxy-1)
   the lexicographic index of pixel (x,y) is 
   index = (y-1)*head.nsx + x-1 where 1 <= index <= head.nsx*head.nsy */

void sir_update_head(struct sirheader *head, float *stval);
/* routine to update old-style header info to include additional
   information in the newer style header.  Sets default anodata
   and v_min v_max values based on image contents.  Call after reading
   image content information. */


/* double precision SIR routine function prototypes */

void dpixtolatlon(double x, double y, double *alon, double *alat, 
		 int iopt, float xdeg, float ydeg,
		 float ascale, float bscale, float a0, float b0);
/* given pixel coordinates x,y and projection parameters iopt..b0
   computes the corresponding lon,lat */

void dlatlon2pix(double alon, double alat, double *x, double *y, 
		 int iopt, float xdeg, float ydeg,
		 float ascale, float bscale, float a0, float b0);
/* given a lon and lat and projection parameters iopt..b0
   computes the corresponding x and y values (which may be outside
   of conventional SIR image */

void df2ipix(double x, double y, int *ix, int *iy, int nsx, int nsy);
/* given pixel coordinates x,y as floats and the image size
   compute the corresponding  integer pixel coordinates.  Note that
   pixel coordinates are (1 <= x <= nsx) (1 <= y <= nsy) the fractional
   part of the float is truncated off, thus, (1,1) corresponds to the
   lower-left corner of the lower-left pixel.  A returned
   ix=0 or iy=0 indicates that pixel is outside of image (out of
   range).  If ix and iy are in range (non-zero) lexicographic index
   of pixel (x,y) is index = (y-1)*nsx + x where 1 <= index <= nsx*nsy */

void dsir_pix2latlon(double x, double y, double *alon, double *alat, sir_head *head);

int dsir_latlon2pix(double alon, double alat, double *x, double *y, sir_head *head);

int dsir_pix(double x, double y, int *ix, int *iy, sir_head *head);
#endif
}

#endif
