#!/usr/bin/env python
""" 
Created Apr 7, 2017

This demo script plots plot a SIR image from a file and adds a box
The code also illustrates how to convert lat/lon to pixel location

@author: DGL
"""
######
# Imports
######
import os, sys
import numpy as np
import matplotlib.pyplot as plt
from sir_io import *
from sir_geom import *

######
# Export function definition
######
def save_it(fname, img_dir='.'):
    """ Helper function to export figures """
    if not os.access(img_dir, os.F_OK):
        os.mkdir(img_dir)
    plt.savefig('{0}/{1}.png'.format(img_dir,fname), dpi=150)
    #plt.savefig('{0}/{1}.pdf'.format(img_dir,fname))


######
# Plot the lat/lon box over the sir
######
def plot_latlon_box(sir_fname, lat_range, lon_range, steps, out_fname):
    """ Plots a box in lat/lon coords over a sir image

    sir_fname: the SIR file name. The SIR image header defines things like the map
        projection and so forth.
    lat_range: a list defined as [min_lat, max_lat]
    lon_range: a list defined as [min_lon, max_lon]
    step: (integer greater than zero) interpolation steps

    out_fname: output filename for exported image with overlaid box

    """
    opts_points = True
    # Load in the SIR image
    sir = loadsir(sir_fname)
    # sir[0] is the image array; sir[1] is the header; sir[2]=iaopt; sir[3]=descript

    print('File name: ',sir_fname)
    printsirhead(sir[1])

    # Convert lat/lon coords to pixels in lower-left and upper-right corners
    pix_ll = latlon2pix(lon_range[0], lat_range[0], sir[1])
    pix_ur = latlon2pix(lon_range[1], lat_range[1], sir[1])
    # And to upper-left and lower-right corners
    pix_ul = latlon2pix(lon_range[0], lat_range[1], sir[1])
    pix_lr = latlon2pix(lon_range[1], lat_range[0], sir[1])
    pix_ll = np.floor(pix_ll)
    pix_ur = np.floor(pix_ur)
    pix_ul = np.floor(pix_ul)
    pix_lr = np.floor(pix_lr)    
    corners = [pix_ll, pix_ul, pix_ur, pix_lr]

    # Draw SIR image
    plt.figure()
    #plt.imshow(sir[0], cmap=plt.cm.gray, interpolation='None', aspect='equal')
    plt.imshow(sir[0], cmap=plt.cm.gray, interpolation='nearest', aspect='equal')
    plt.hold(True)

    # Draw corners
    if opts_points:
        for corn in corners:
            plt.plot(corn[0], corn[1], 'xr', lw=2)

#        print(lon_range,lat_range)
#        for corn in corners: # check corners
#            latlon = pix2latlon(corn[0], corn[1], sir[1])
#            print(np.array(latlon))

    # plot Overlay box
    if steps == 0:  # simple rectangular box
        corn_array = np.vstack((corners, corners[0]))
        plt.plot(corn_array[:,0], corn_array[:,1], 'b', lw=2)
    elif steps > 0:     # generate interpolated box
        dlon = (lon_range[1]-lon_range[0])/steps
        dlat = (lat_range[1]-lat_range[0])/steps
        n = 0
        llist = [(0,0)]*(steps*4+1)
        for i in range(steps):
            llist[n] = latlon2pix(lon_range[0]+i*dlon, lat_range[0], sir[1])
            n +=1
        for i in range(steps):
            llist[n] = latlon2pix(lon_range[1], lat_range[0]+i*dlat, sir[1])
            n +=1
        for i in range(steps):
            llist[n] = latlon2pix(lon_range[1]-i*dlon, lat_range[1], sir[1])
            n +=1
        for i in range(steps):
            llist[n] = latlon2pix(lon_range[0], lat_range[1]-i*dlat, sir[1])
            n +=1
        llist[n] = latlon2pix(lon_range[0], lat_range[0], sir[1])
        llist=np.array(llist)
        plt.plot(llist[:,0], llist[:,1], 'g', lw=1)

    # Clean up figure
    plt.axis('image')
    plt.title(sir_fname)
    plt.xlabel('x')
    plt.ylabel('y')

    # Export figure
#    save_it(out_fname)
#    plt.close()

def main():
    # A test driver that illustrates plotting an arbitrary
    # lat/lon box on a sir image
    sir_fname = 'greeni.sir'
    #sir_fname = 'queh-a-E2N04-10-10.sir'
    out_fname = './' + sir_fname
    lat_range = [70., 74.0]
    lon_range = [-46., -36.]
    steps = 3
    plot_latlon_box(sir_fname, lat_range, lon_range, steps, out_fname)

    save_it(out_fname)  # generate .png

    plt.show()

    if 1==1:  # show .png file
        cmd='pwd; ls * ; eog ' + out_fname +'.png'
        print(cmd)
        os.system(cmd)

if __name__ == '__main__':
    main()
